package ca.datasports.sb.webservicelistener;

import javax.jws.WebMethod;
import javax.jws.WebService;
import javax.xml.ws.BindingType;
import javax.xml.ws.Endpoint;

/**
 * This class exposes the public WebService methods, and relays them back to its parent
 * via the {@link IServiceListener} interface.  It also encapsulates all logic related
 * to the management of the WebService.
 * 
 * @author Phil Martin, <a href="mailto:phil.martin@datasports.ca">phil.martin@datasports.ca</a>.
 * @see IServiceListener
 * @see WebServiceListener
 */
@WebService
@BindingType(value="http://java.sun.com/xml/ns/jaxws/2003/05/soap/bindings/HTTP/")
public class ServiceImplementation {

	private IServiceListener _listener = null;
	private String _serviceUri = "";
	private Endpoint _serviceEndpoint = null;

	/**
	 * For some reason, wsgen fails if it can't find this constructor.  It is not
	 * to be used by application code. Use the parameterized constructor instead.
	 * @deprecated
	 * @see #ServiceImplementation(IServiceListener, String)
	 * @throws Exception
	 */
	@Deprecated
	public ServiceImplementation() throws Exception {
		throw new Exception("Don't use, this is a stub required by wsgen");
	}
	
	/**
	 * Use this constructor, otherwise the necessary member variables will not
	 * be set.
	 * 
	 * @param listener an object that implements {@link IServiceListener}, which will provide
	 * the actual implementation of the WebService methods exposed by this wrapper class. 
	 * @param serviceUri the URI at which the WebService should be exposed.
	 * @see IServiceListener
	 * @see WebServiceListener
	 */
	public ServiceImplementation(IServiceListener listener, String serviceUri) {
		_listener = listener;
		_serviceUri = serviceUri;
	}

	/**
	 * Uses the {@link Endpoint} class to publish the WebService at the URI specified
	 * in the constructor.
	 * 
	 * @see Endpoint
	 */
	@WebMethod(exclude=true)
	public void start()
	{
		_serviceEndpoint = Endpoint.publish(_serviceUri, this);
	}

	/**
	 * Stops the running WebService if it has been started, and cleans up the 
	 * {@link Endpoint} instance.
	 * 
	 * @see Endpoint
	 */
	@WebMethod(exclude=true)
	public void stop()
	{
		if (_serviceEndpoint != null && _serviceEndpoint.isPublished()) {
			_serviceEndpoint.stop();
		}
		_serviceEndpoint = null;
	}

	/**
	 * Public WebMethod exposed by this WebService.
	 * <p>
	 * This call will be relayed to the object's listener via the {@link IServiceListener} interface.
	 * 
	 * @param notificationMessage
	 * @return Condition indicating success or failure.
	 * @see IServiceListener
	 * @see WebServiceListener
	 */
	@WebMethod
	public boolean NewMessage(String notificationMessage)
	{
		try {
			return _listener.NewMessage(notificationMessage);
		}
		catch (Exception exc) {
			return false;
		}
	}

	/**
	 * Public WebMethod exposed by this WebService.
	 * <p>
	 * This call will be relayed to the object's listener via the {@link IServiceListener} interface.
	 * 
	 * @param key
	 * @param value
	 * @return Condition indicating success or failure.
	 * @see IServiceListener
	 * @see WebServiceListener
	 */
	@WebMethod
	public boolean SetValue(String key, String value)
	{
		try {
			return _listener.SetValue(key, value);
		}
		catch (Exception exc) {
			return false;
		}
	}
}
