package ca.datasports.sb.webservicelistener;


import org.slf4j.Logger;

import com.streambase.sb.*;
import com.streambase.sb.adapter.*;
import com.streambase.sb.operator.*;

import java.net.InetAddress;
import ca.datasports.sb.utilities.SchemaValidator;

/**
 * Datasports WebServiceListener Adapter.
 * <p>
 * Complete documentation can be found at 
 * <a href="http://webservicelistener.datasports.ca">http://webservicelistener.datasports.ca</a>.
 * For other information about this adapter or about Datasports, contact 
 * <a href="mailto:info@datasports.ca">info@datasports.ca</a>.
 * <p>
 * This class, like all input adapters, extends the InputAdapter class.  In addition
 * to the StreamBase-specific interface Parameterizable and Java's Runnable, it implements
 * the custom {@link IServiceListener} interface, used to provide the implementation of the web service
 * exposed by the {@link ServiceImplementation} class.
 * <p>
 * This example makes use of the JAX-WS 2.0 library to implement web services and clients.
 * JAX-WS 2.0 documentation can be found at <a href="http://java.sun.com/developer/technicalArticles/J2SE/jax_ws_2/">
 * http://java.sun.com/developer/technicalArticles/J2SE/jax_ws_2/</a>   
 * 
 * @author Phil Martin, <a href="mailto:phil.martin@datasports.ca">phil.martin@datasports.ca</a>.
 * 
 */
public class WebServiceListener extends InputAdapter implements
		Parameterizable, Runnable, IServiceListener {

	public static final long serialVersionUID = 1277744957948L;
	private static final String displayName = "Web Service Listener";
	private Logger logger;

	// Member variables behind the public properties
	private boolean _controlPort;	// Whether or not to expose a control port for run-time changing of adapter properties.
	private boolean _enabled;		// Should the adapter run or not.
	private String _baseUri;		// URI at which the service should run (the name of the local machine will be prepended). 
	
	private boolean _inited = false;	// Tracks state for proper handling of resume() callback.
	
	// Constants for ports and schema definitions.
	private static final int INPUT_PORT_COUNT = 1;
	private static final int OUTPUT_PORT_COUNT = 2;
	
	// Input port and schema details 
	private static final int ADAPTER_CONTROL_PORT = 0;
	private Schema adapterControlSchema;
	private static class ADAPTER_CONTROL_SCHEMA 
	{
		public static String Enabled = "Enabled";
		public static String BaseUri = "BaseUri";
	}

	// Output port and schema details
	private static final int MESSAGE_EVENT_PORT = 0;
	private Schema messageEventSchema;
	private static class MESSAGE_EVENT_SCHEMA
	{
		public static String Message = "Message";
	}

	private static final int KEY_VALUE_EVENT_PORT = 1;
	private Schema keyValueEventSchema;
	private static class KEY_VALUE_EVENT_SCHEMA
	{
		public static String Key = "Key";
		public static String Value = "Value";
	}

	private ServiceImplementation _serviceImpl = null;	// This will hold an instance of the class actually exposed as a web service.
	
	/**
	 * The constructor is called when the Adapter instance is created, but before the Adapter 
	 * is connected to the StreamBase application.
	 * <p>
	 * Here we set some properties relevant to StreamBase Studio, such as the display name 
	 * of the adapter, hints about the number of input and output ports, and initial values
	 * for the adapter's properties.
	 * <p>
	 * We also create our input and output schemas.
	 *  
	 */
	public WebServiceListener() {
		super();
		logger = getLogger();
		setDisplayName(displayName);
		setShortDisplayName(this.getClass().getSimpleName());
		setControlPort(false);
		setEnabled(true);
		setBaseUri(":8080/WebServiceListener");

		setPortHints(0, OUTPUT_PORT_COUNT);
		
		adapterControlSchema = new Schema( "WebServiceListenerControlSchema", new Schema.Field[] {
				Schema.createField(DataType.BOOL, ADAPTER_CONTROL_SCHEMA.Enabled),
				Schema.createField(DataType.STRING, ADAPTER_CONTROL_SCHEMA.BaseUri)
			});

		messageEventSchema = new Schema("WebServiceListenerMessageEventSchema", new Schema.Field[] {
			Schema.createField(DataType.STRING, MESSAGE_EVENT_SCHEMA.Message),
		});
	
		keyValueEventSchema = new Schema("WebServiceListenerKeyValueEventSchema", new Schema.Field[] {
				Schema.createField(DataType.STRING, KEY_VALUE_EVENT_SCHEMA.Key),
				Schema.createField(DataType.STRING, KEY_VALUE_EVENT_SCHEMA.Value),
			});
	
	}

	/**
	 * Typecheck this adapter. The typecheck method is called after the adapter instance is
	 * connected in the StreamBase application. The adapter should validate its parameters 
	 * and throw TypecheckException if any problems are found.
	 * <p>
	 * Here we set the number of input ports, validate that the input port has the correct
	 * schema if it is enabled, and set the output schemas.
	 * 
	 */
	public void typecheck() throws TypecheckException {

		if (_controlPort) {
			requireInputPortCount(INPUT_PORT_COUNT);
			SchemaValidator.validateSchema(getInputSchema(ADAPTER_CONTROL_PORT), adapterControlSchema);
		}
		else {
			requireInputPortCount(0);
		}
		
		setOutputSchema(MESSAGE_EVENT_PORT, messageEventSchema);
		setOutputSchema(KEY_VALUE_EVENT_PORT, keyValueEventSchema);
	}

	/**
	 * Initialize the adapter. If typecheck succeeds, the init method is called before
	 * the StreamBase application is started. Note that your adapter is not required to
	 * define the init method, unless you need to register a runnable or perform
	 * initialization of a resource such as, for example, a JDBC pool.
	 */
	public void init() throws StreamBaseException {
		super.init();

		// Register the object so it will be run as a thread managed by StreamBase.
		registerRunnable(this, true);
	}

	/**
	 * Shutdown adapter by cleaning up any resources used. When the application is shutting 
	 * down, the adapter's shutdown method will be called first. Once this has returned, all 
	 * threads should exit and the adapter is considered shutdown.
	 * <p>
	 * For this adapter, the only allocated resources are in the JAX-WS subsystem, which are
	 * released in the {@link #stopListening()} method.
	 */
	public void shutdown() {
		stopListening();
	}

	/**
	 * Main thread of the adapter. At this point, the application begins to run.
	 * StreamBase will start threads for any managed runnables registered by
	 * earlier calls to registerRunnable.
	 * <p>
	 * In this method, we kick off the processing of input events (in this case,
	 * by calling {@link #startListening()}).
	 * 
	 */
	public void run() {
		_inited = true;
		if (shouldRun())
		{
			startListening();
		}
		
		try {
			while (shouldRun())
			{
				Thread.sleep(3000);
			}
		} catch (InterruptedException e) {
			getLogger().debug("sleep interrupted, shutting down.");
		}
	}

	/**
	 * Here we initialize an instance of the {@link ServiceImplementation} class,
	 * starting the WebService at the specified address.
	 * 
	 * @see ServiceImplementation
	 */
	private void startListening() {
		String serviceUri = "";
		
		try {
			if (_enabled) {
				serviceUri = "http://" + InetAddress.getLocalHost().getHostName() + _baseUri;
				if (_serviceImpl != null) {
					stopListening();
				}
				logger.info("Starting Web Service at URI: " + serviceUri);
				
				_serviceImpl = new ServiceImplementation(this, serviceUri);
				_serviceImpl.start();

			}
		}
		catch (Exception exc) {
			logger.error("Failure starting Web Service at uri '" + serviceUri + "': " + exc.getMessage());
		}
	}

	/**
	 * Here we shut down the listening WebService.  This would be done when 
	 * the adapter itself is shut down, suspended, or disabled, or when 
	 * properties have changed via the input control port which requires 
	 * an existing running WebService instance to be stopped so that the 
	 * new instance can be started with the newly-specified parameters.
	 * 
	 * @see ServiceImplementation
	 */
	private void stopListening() {
		try {
			if (_serviceImpl != null) {
				logger.info("Stopping Web Service.");
				_serviceImpl.stop();
			}
		}
		catch (Exception exc) {
			logger.error("Failure stopping service: '" + exc.getMessage() + "'");
		}
		finally {
			_serviceImpl = null;
		}
	}

	/**
	 * This method will be called when an adapter starts or resumes execution,
	 * before any registered runnables are started or resumed.  Note that if the
	 * application is shutdown directly from a suspended state, this will
	 * <b>not</b> be called; instead {@link #shutdown()} will be called.
	 *
	 * resume() is a callback that is called by the StreamBase runtime.
	 */
	public void resume() {
		if (_inited) {
			startListening();
		}
	}

	/**
	 * This method will be called when an adapter suspends, before any
	 * registered runnables are suspended.
	 * <p>
	 * suspend() is a callback that is called by the StreamBase runtime.
	 * If this is called, we invoke {@link #stopListening()}, which contains
	 * the logic to release JAX-WS resources and stop listening for remote calls.
	 */
	public void suspend() {
		stopListening();
	}


	/***************************************************************************************
	 * The getter and setter methods provided by the Parameterizable object.               *
	 * StreamBase Studio uses them to determine the name and type of each property         *
	 * and obviously, to set and get the property values.                                  *
	 ***************************************************************************************/

	public void setControlPort(boolean ControlPort) {
		_controlPort = ControlPort;
	}
	
	public boolean getControlPort() {
		return _controlPort;
	}
	
	public void setEnabled(boolean Enabled) {
		_enabled = Enabled;
	}

	public boolean getEnabled() {
		return _enabled;
	}

	public void setBaseUri(String BaseUri) {
		_baseUri = BaseUri;
	}

	public String getBaseUri() {
		return _baseUri;
	}

	/**
	 * Implements the IServiceListener.NewMessage() method, which 
	 * will be called by the {@link ServiceImplementation} instance when its
	 * NewMessage() method is called.
	 * <p>
	 * For testing purposes, this method will fail with an exception if
	 * the message contains the token "FAIL" (case-insensitive comparison).
	 * 
	 * @param newMessage the message value to process
	 * @see IServiceListener
	 * @see ServiceImplementation
	 */
	@Override
	public boolean NewMessage(String newMessage) throws Exception {
		if (newMessage.toUpperCase().contains("FAIL")) {
			throw new Exception("Failure in processing message.");
		}
		SendMessageTuple(newMessage);
		return true;
	}

	/**
	 * Implements the IServiceListener.SetValue() method, which 
	 * will be called by the {@link ServiceImplementation} instance when its
	 * NewMessage() method is called.
	 * <p>
	 * For testing purposes, this method will fail with an exception if
	 * the key contains the token "FAIL" (case-insensitive comparison).
	 * 
	 * @param key specifies the <b>key</b> to be passed through the output adapter as part of a Key/Value event.
	 * @param value specifies the <b>value</b> to be passed through the output adapter as part of a Key/Value event.
	 * @see IServiceListener
	 * @see ServiceImplementation
	 */
	@Override
	public boolean SetValue(String key, String value) throws Exception {
		if (key.toUpperCase().contains("FAIL")) {
			throw new Exception("Failure in setting value.");
		}
		SendKeyValueTuple(key, value);
		return true;
	}

	/**
	 * Composes a tuple to represent the call to the NewMessage() method on
	 * the running WebService, and sends it out the appropriate output port.
	 * 
	 * @param message received from the remote caller.
	 * @throws Exception
	 */
	private void SendMessageTuple(String message) throws Exception {
		try {
			Tuple messageTuple = getOutputSchema(MESSAGE_EVENT_PORT).createTuple();
			messageTuple.setField(MESSAGE_EVENT_SCHEMA.Message, message);

			sendOutput(MESSAGE_EVENT_PORT, messageTuple);
		}
		catch (Exception exc) {
			throw new Exception("SendMessageTuple failed, " + exc.getMessage());
		}
	}
	
	/**
	 * Composes a tuple to represent the call to the NewMessage() method on
	 * the running WebService, and sends it out the appropriate output port.
	 * 
	 * @param key received from the remote caller.
	 * @param value received from the remote caller.
	 * @throws Exception
	 */
	private void SendKeyValueTuple(String key, String value) throws Exception {
		try {
			logger.info("in SendKeyValueTuple: key = " + key + ", value = " + value);
			Tuple keyValueTuple = getOutputSchema(KEY_VALUE_EVENT_PORT).createTuple();
			keyValueTuple.setField(KEY_VALUE_EVENT_SCHEMA.Key, key);
			keyValueTuple.setField(KEY_VALUE_EVENT_SCHEMA.Value, value);
			
			sendOutput(KEY_VALUE_EVENT_PORT, keyValueTuple);
		}
		catch (Exception exc) {
			throw new Exception("SendKeyValueTuple failed, " + exc.getMessage());
		}
	}
	
	/**
	 * Implement this method if you have an input port on your adapter.
	 * In many cases, an input adapter may not require an input port, since
	 * it is listening for events external to StreamBase.  The Datasports
	 * WebServiceListener adapter uses a property called ControlPort to indicate
	 * if this optional input port should be enabled or not.
	 * <p>
	 * This port is used to change the values of the properties of the adapter,
	 * which is a common usage for an input port on an input adapter.  Properties
	 * that are to be left unchanged may be set to null.  The new values are
	 * applied immediately via the call to {@link #startListening()} or 
	 * {@link #stopListening()}.
	 * 
	 */
	@Override
	public void processTuple(int inputPort, Tuple input) throws StreamBaseException	{
		if (!input.isNull(ADAPTER_CONTROL_SCHEMA.Enabled)) {
			setEnabled(input.getBoolean(ADAPTER_CONTROL_SCHEMA.Enabled));
		}
		
		if (!input.isNull(ADAPTER_CONTROL_SCHEMA.BaseUri)) {
			setBaseUri(input.getString(ADAPTER_CONTROL_SCHEMA.BaseUri));
		}

		if (_enabled) {
			startListening();
		}
		else {
			stopListening();
		}
	}
	
	/**
	 * Called by StreamBase Studio to determine how many input and output ports
	 * the adapter has.  We override this method because the optional input
	 * port means that the number of input ports may change at design time.
	 *  
	 */
	@Override
	public PortCounts getPortCounts() throws TypecheckException {
		PortCounts counts = new PortCounts(
				_controlPort ? INPUT_PORT_COUNT : 0, 
				OUTPUT_PORT_COUNT);
		return counts;
	}
	
}
